import copy
import time
from typing import Any
from uuid import UUID

from ..config.configs import ClientInfo
from ..lang.common_models import Metric
from ..lang.datamodel_enum import ReportTypeEnum
from ..reports.ml_report_model import (
    FlattenMetric,
    GeneralApplicationReportModel,
    MicroserviceInstance,
)
from .abstract_report import AbstractReport


class GeneralApplicationReport(AbstractReport):
    """
    GeneralApplicationReport manages the reporting of application metrics and inference data.

    Parameters
    ----------
    client_config : ClientInfo
        Configuration settings related to the client.

    Attributes
    ----------
    client_config : ClientInfo
        A deep copy of the client configuration.
    init_time : float
        The initialization time of the report.
    report : GeneralApplicationReportModel
        The current state of the report.
    execution_instance : MicroserviceInstance
        An instance representing the current execution context.
    previous_reports : list[MicroserviceInstance]
        A list of previous execution instances.

    Methods
    -------
    reset() -> None
        Reset the report to an initial state.
    process_previous_report(previous_report_dict: dict) -> None
        Process and incorporate a previous report.
    observe_metric(report_type: ReportTypeEnum, stage: str, metric: Metric) -> None
        Observe and record a metric.
    observe_inference(inference_value: Any) -> None
        Observe and record inference data.
    observe_inference_metric(metric: Metric) -> None
        Observe and record an inference-specific metric.
    """

    def __init__(self, client_config: ClientInfo) -> None:
        """
        Initialize an instance of GeneralApplicationReport.

        Parameters
        ----------
        client_config : ClientInfo
            Configuration settings related to the client.
        """
        self.client_config = copy.deepcopy(client_config)
        self.reset()
        self.init_time = time.time()

    def reset(self) -> None:
        """
        Reset the report to an initial state.

        Notes
        -----
        - This method initializes a new report model and sets up the execution instance and previous reports list.
        """
        self.report = GeneralApplicationReportModel()
        self.execution_instance = MicroserviceInstance(
            id=UUID(self.client_config.instance_id),
            name=self.client_config.name,
            functionality=self.client_config.functionality,
            stage=self.client_config.stage_id,
        )
        self.previous_reports: list[MicroserviceInstance] = []

    def process_previous_report(self, previous_report_dict: dict) -> None:
        """
        Process and incorporate a previous report.

        Parameters
        ----------
        previous_report_dict : dict
            Dictionary representation of a previous report.

        Notes
        -----
        - This method assumes the last metric in the previous report was observed by the previous instance.
        - It appends the metrics from the previous report to the current report.
        """
        previous_report = GeneralApplicationReportModel(**previous_report_dict)
        self.previous_reports.append(previous_report.metrics[-1].instance)
        for metric in previous_report.metrics:
            self.report.metrics.append(metric)

    def observe_metric(
        self, report_type: ReportTypeEnum, stage: str, metric: Metric
    ) -> None:
        """
        Observe and record a metric.

        Parameters
        ----------
        report_type : ReportTypeEnum
            The type of report being generated.
        stage : str
            The stage of the process in which the metric is recorded.
        metric : Metric
            The metric to be recorded.
        """
        flatten_metric = FlattenMetric(
            metric_name=metric.metric_name,
            records=metric.records,
            unit=metric.unit,
            description=metric.description,
            stage=stage,
            report_type=report_type,
            instance=self.execution_instance,
            previous_instances=self.previous_reports,
        )
        self.report.metrics.append(flatten_metric)

    def observe_inference(self, inference_value: Any) -> None:
        """
        Observe and record inference data.

        Parameters
        ----------
        inference_value : Any
            The value of the inference to be recorded.

        Notes
        -----
        - This method records inference values as a metric with the name "Inference" and report type ml_specific.
        """
        flatten_metric = FlattenMetric(
            metric_name="Inference",
            records=inference_value,
            stage=self.client_config.stage_id,
            report_type=ReportTypeEnum.ml_specific,
            instance=self.execution_instance,
            previous_instances=self.previous_reports,
        )
        self.report.metrics.append(flatten_metric)

    def observe_inference_metric(self, metric: Metric) -> None:
        """
        Observe and record an inference-specific metric.

        Parameters
        ----------
        metric : Metric
            The inference-specific metric to be recorded.
        """
        flatten_metric = FlattenMetric(
            metric_name=metric.metric_name,
            records=metric.records,
            unit=metric.unit,
            description=metric.description,
            stage=self.client_config.stage_id,
            report_type=ReportTypeEnum.ml_specific,
            instance=self.execution_instance,
            previous_instances=self.previous_reports,
        )
        self.report.metrics.append(flatten_metric)



import pickle
import socket
from typing import Callable

from ..config.configs import SocketCollectorConfig
from .base_collector import BaseCollector


class SocketCollector(BaseCollector):
    """
    SocketCollector handles the collection of data over a TCP socket.

    Parameters
    ----------
    config : SocketCollectorConfig
        Configuration settings for the socket collector.
    process_report : Callable
        A callable function to process incoming reports.

    Attributes
    ----------
    config : SocketCollectorConfig
        The socket collector configuration.
    host : str
        The hostname or IP address to bind the socket.
    port : int
        The port number to bind the socket.
    backlog : int
        The maximum number of queued connections.
    bufsize : int
        The maximum size of data to be received at once.
    process_report : Callable
        A function to process the received report.
    execution_flag : bool
        Flag to control the execution loop.

    Methods
    -------
    start_collecting()
        Start the socket server to collect and process incoming data.
    """

    def __init__(self, config: SocketCollectorConfig, process_report: Callable) -> None:
        """
        Initialize an instance of SocketCollector.

        Parameters
        ----------
        config : SocketCollectorConfig
            Configuration settings for the socket collector.
        process_report : Callable
            A callable function to process incoming reports.
        """
        self.config = config
        self.host = config.host
        self.port = config.port
        self.backlog = config.backlog
        self.bufsize = config.bufsize
        self.process_report = process_report
        self.execution_flag = True

    def start_collecting(self) -> None:
        """
        Start the socket server to collect and process incoming data.

        Notes
        -----
        - This method starts a TCP socket server that listens for incoming connections.
        - Data received from clients is deserialized using pickle and then processed using the `process_report` function.
        - The server runs indefinitely until the `execution_flag` is set to False.
        """
        server_socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        server_socket.bind((self.host, self.port))
        server_socket.listen(self.backlog)

        while self.execution_flag:
            client_socket, _ = server_socket.accept()
            data = b""
            while True:
                packet = client_socket.recv(self.bufsize)
                if not packet:
                    break
                data += packet

            report = pickle.loads(data)
            self.process_report(report)
            client_socket.close()



import asyncio
import time

import docker
from docker.models.containers import Container

from qoa4ml.reports.resources_report_model import (
    DockerContainerMetadata,
    DockerContainerReport,
    ResourceReport,
)

BYTES_TO_MB = 1024.0 * 1024.0


async def get_container_stats(
    container: Container,
) -> DockerContainerReport:
    container_stats = {}
    container_stats["cpu"] = {}
    container_stats["mem"] = {}

    timestamp = time.time()
    stat = await asyncio.to_thread(container.stats, stream=False)

    usage_delta = (
        stat["cpu_stats"]["cpu_usage"]["total_usage"]
        - stat["precpu_stats"]["cpu_usage"]["total_usage"]
    )
    system_delta = (
        stat["cpu_stats"]["system_cpu_usage"] - stat["precpu_stats"]["system_cpu_usage"]
    )
    len_cpu = stat["cpu_stats"]["online_cpus"]
    cpu_percentage = (usage_delta / system_delta) * len_cpu * 100
    container_stats["cpu"] = {"cpu_percentage": cpu_percentage}

    # Bytes to Mb
    container_stats["mem"] = {
        "memory_usage": stat["memory_stats"]["usage"] / BYTES_TO_MB
    }

    container_id = container.id
    container_image = container.image
    if not container_id or not container_image:
        raise RuntimeError("container id or container image is None")
    return DockerContainerReport(
        metadata=DockerContainerMetadata(
            id=container_id, image=container_image.tags[0]
        ),
        timestamp=timestamp,
        cpu=ResourceReport(usage={"cpu_percentage": cpu_percentage}),
        mem=ResourceReport(
            usage={"memory_usage": stat["memory_stats"]["usage"] / BYTES_TO_MB}
        ),
    )


async def get_all_container_stats(client):
    tasks = []
    for container in client.containers.list():
        if container.status == "running":
            tasks.append(get_container_stats(container))
    results = await asyncio.gather(*tasks)
    return results


async def get_container_list_stats(
    client: docker.DockerClient, container_list: list[str]
):
    tasks = []
    for container_name in container_list:
        container = client.containers.get(container_name)
        if container.status == "running":
            tasks.append(get_container_stats(container))
    results = await asyncio.gather(*tasks)
    return results


def get_docker_stats(
    client: docker.DockerClient, container_list: list[str]
) -> list[DockerContainerReport]:
    if container_list:
        return asyncio.run(get_container_list_stats(client, container_list))

    return asyncio.run(get_all_container_stats(client))



import math
import time
from datetime import datetime
from pathlib import Path

from tinyflux import Point, TimeQuery, TinyFlux


class EmbeddedDatabase:
    def __init__(self, db_path: Path) -> None:
        self.db = TinyFlux(db_path)

    def insert(self, timestamp: float, tags: dict, fields: dict):
        timestamp_datetime = datetime.fromtimestamp(timestamp)
        datapoint = Point(time=timestamp_datetime, tags=tags, fields=fields)
        self.db.insert(datapoint, compact_key_prefixes=True)

    def get_lastest_timestamp(self):
        time_query = TimeQuery()
        timestamp = datetime.fromtimestamp(math.floor(time.time()))
        return self.db.search(time_query >= timestamp)



from typing import TYPE_CHECKING

import lazy_import

from ..collector.host_object import HostObject
from ..config.configs import MQTTConnectorConfig
from ..utils.qoa_utils import qoaLogger
from .base_connector import BaseConnector

if TYPE_CHECKING:
    import paho.mqtt.client as mqtt
else:
    mqtt = lazy_import.lazy_module("paho.mqtt")


# TODO: this client handle both connector and collector
class MqttConnector(BaseConnector):
    # This class will handle all the mqtt connection for each client application
    # FIX: what is host object?
    def __init__(self, host_object: HostObject, configuration: MQTTConnectorConfig):
        # from paho.mqtt.client import Client as MqttClient
        # from paho.mqtt.enums import CallbackAPIVersion
        # Init the host object to return message
        self.host_object = host_object
        # Init the send/receive queue
        self.pub_queue = configuration.in_queue
        self.sub_queue = configuration.out_queue
        # Create the mqtt client
        self.test = mqtt
        self.client = mqtt.Client(
            callback_api_version=mqtt.CallbackAPIVersion.VERSION2,
            client_id=configuration.client_id,
            clean_session=False,
            userdata=None,
            transport="tcp",
        )
        # Set some functional method
        self.client.on_connect = self.on_connect
        self.client.on_message = self.on_message
        # Connect to mqtt broker
        self.client.connect(
            configuration.broker_url,
            configuration.broker_port,
            configuration.broker_keepalive,
        )

    def on_connect(self, client, userdata, flags, rc):
        qoaLogger.debug("Connected with result code " + str(rc))
        # Subscribing in on_connect() means that if we lose the connection and
        # reconnect then subscriptions will be renewed.
        client.subscribe(self.sub_queue)

    def on_message(self, client, userdata, msg):
        # Pass the data to the host object
        self.host_object.message_processing(client, userdata, msg)

    def stop(self):
        # stop the connection
        self.client.disconnect()

    def start(self):
        # Start looking for data from broker
        self.client.loop_start()

    def send_data(self, body_message: str):
        # Send data in form of text message
        self.client.publish(self.pub_queue, body_message)



import json

from devtools import debug

from ..config.configs import DebugConnectorConfig
from .base_connector import BaseConnector


class DebugConnector(BaseConnector):
    """
    DebugConnector provides a connector for debugging purposes, enabling the logging of messages.

    Parameters
    ----------
    config : DebugConnectorConfig
        Configuration settings for the debug connector.

    Attributes
    ----------
    silence : bool
        Flag to suppress debugging output if set to True.

    Methods
    -------
    send_report(body_message: str) -> None
        Send and debug the message.
    """

    def __init__(self, config: DebugConnectorConfig):
        """
        Initialize an instance of DebugConnector.

        Parameters
        ----------
        config : DebugConnectorConfig
            Configuration settings for the debug connector.
        """
        self.silence = config.silence

    def send_report(self, body_message: str) -> None:
        """
        Send and debug the message.

        Parameters
        ----------
        body_message : str
            The message body to be sent and debugged.

        Notes
        -----
        If `silence` is set to False, the message will be logged for debugging purposes.
        """
        if not self.silence:
            debug(json.loads(body_message))

    def check_connection(self) -> bool:
        return True



from __future__ import annotations

import math
import time
from abc import ABC, abstractmethod
from typing import Any

from ..config.configs import ClientInfo, ProbeConfig
from ..connector.base_connector import BaseConnector
from ..utils.qoa_utils import make_folder
from ..utils.repeated_timer import RepeatedTimer


class Probe(ABC):
    def __init__(
        self,
        config: ProbeConfig,
        connector: BaseConnector,
        client_info: ClientInfo | None = None,
    ) -> None:
        self.config = config
        self.client_info = client_info
        self.frequency = self.config.frequency
        self.monitoring_interval = 1.0 / self.frequency
        self.execution_flag = False
        self.report_thread = None
        self.log_latency_flag = self.config.log_latency_flag
        if self.log_latency_flag:
            self.latency_logging_path = config.latency_logging_path
            make_folder(self.latency_logging_path)
        self.max_latency = 0.0
        self.connector = connector

    @abstractmethod
    def create_report(self) -> Any:
        pass

    def reporting(self):
        report = self.create_report()
        self.connector.send_report(report)

    def start_reporting(self, background: bool = True):
        """
        background = False for blocking reporting
        """
        self.execution_flag = True
        current_time = time.time()
        time.sleep(math.ceil(current_time) - current_time)
        self.timer = RepeatedTimer(self.monitoring_interval, self.reporting)
        if not background:
            self.timer.thread.join()

    def stop_reporting(self):
        if not hasattr(self, "timer"):
            raise RuntimeError("Can't stop reporting when the timer is not created yet")
        self.timer.stop()

    def send_report(self, report):
        start = time.time()
        self.connector.send_report(report)
        if self.log_latency_flag and self.latency_logging_path:
            self.write_log(
                (time.time() - start) * 1000,
                self.latency_logging_path + "report_latency.txt",
            )

    def write_log(self, latency, filepath: str):
        with open(filepath, "a", encoding="utf-8") as file:
            file.write(str(latency) + "\n")



from .pynvml_forked import (
    nvmlDeviceGetCount,
    nvmlDeviceGetHandleByIndex,
    nvmlDeviceGetMaxClockInfo,
    nvmlDeviceGetMemoryInfo,
    nvmlDeviceGetNumGpuCores,
    nvmlDeviceGetUtilizationRates,
    nvmlInit,
)

HAS_NVIDIA_GPU = True
try:
    nvmlInit()
except Exception:
    HAS_NVIDIA_GPU = False


def get_sys_gpu_usage():
    usage = {}
    if not HAS_NVIDIA_GPU:
        return {}
    try:
        device_count = nvmlDeviceGetCount()
    except Exception:
        nvmlInit()
        device_count = nvmlDeviceGetCount()

    for i in range(device_count):
        handle = nvmlDeviceGetHandleByIndex(i)
        util = nvmlDeviceGetUtilizationRates(handle)
        mem = nvmlDeviceGetMemoryInfo(handle)
        usage[f"device_{i+1}_core"] = util.gpu
        usage[f"device_{i+1}_mem"] = mem.used / 1024.0 / 1024
    return usage


def get_sys_gpu_metadata():
    metadata = {}
    if not HAS_NVIDIA_GPU:
        return {}
    try:
        device_count = nvmlDeviceGetCount()
    except Exception:
        nvmlInit()
        device_count = nvmlDeviceGetCount()

    for i in range(device_count):
        handle = nvmlDeviceGetHandleByIndex(i)
        cores = nvmlDeviceGetNumGpuCores(handle)
        clock = nvmlDeviceGetMaxClockInfo(handle, 0)
        mem = nvmlDeviceGetMemoryInfo(handle)
        metadata[f"device_{i+1}"] = {
            "frequency": {"value": clock, "unit": "MHz"},
            "core": cores,
            "mem": {"capacity": mem.total / 1024.0 / 1024, "unit": "Gb"},
        }
    return metadata



import copy
import time
from typing import Any, Optional
from uuid import UUID, uuid4

from ..config.configs import ClientInfo
from ..lang.common_models import Metric
from ..lang.datamodel_enum import ReportTypeEnum
from ..reports.ml_report_model import (
    BaseReport,
    GeneralMlInferenceReport,
    InferenceInstance,
    StageReport,
)
from .abstract_report import AbstractReport


class MLReport(AbstractReport):
    """
    MLReport manages the reporting of machine learning metrics and inference data.

    Parameters
    ----------
    client_config : ClientInfo
        Configuration settings related to the client.

    Attributes
    ----------
    client_config : ClientInfo
        A deep copy of the client configuration.
    init_time : float
        The initialization time of the report.
    previous_report : list[GeneralMlInferenceReport]
        A list of previously processed reports.
    report : GeneralMlInferenceReport
        The current state of the report.

    Methods
    -------
    reset() -> None
        Reset the report to an initial state.
    combine_stage_report(current_stage_report: dict[str, StageReport], previous_stage_report: dict[str, StageReport]) -> dict[str, StageReport]
        Combine metrics from the current and previous stage reports.
    process_previous_report(previous_report_dict: dict) -> None
        Process and incorporate a previous report.
    observe_metric(report_type: ReportTypeEnum, stage: str, metric: Metric) -> None
        Observe and record a metric.
    observe_inference(inference_value: Any) -> None
        Observe and record inference data.
    observe_inference_metric(metric: Metric) -> None
        Observe and record an inference-specific metric.
    generate_report(reset: bool = True, corr_id: Optional[str] = None) -> BaseReport
        Generate the report and optionally reset the current report state.
    """

    def __init__(self, client_config: ClientInfo) -> None:
        """
        Initialize an instance of MLReport.

        Parameters
        ----------
        client_config : ClientInfo
            Configuration settings related to the client.
        """
        self.client_config = copy.deepcopy(client_config)
        self.reset()
        self.init_time = time.time()

    def reset(self) -> None:
        """
        Reset the report to an initial state.

        Notes
        -----
        - This method initializes a new report model and clears the list of previous reports.
        """
        self.previous_report: list[GeneralMlInferenceReport] = []
        self.report = GeneralMlInferenceReport()

    def combine_stage_report(
        self,
        current_stage_report: dict[str, StageReport],
        previous_stage_report: dict[str, StageReport],
    ) -> dict[str, StageReport]:
        """
        Combine metrics from the current and previous stage reports.

        Parameters
        ----------
        current_stage_report : dict
            The current stage report containing metrics.
        previous_stage_report : dict
            The previous stage report containing metrics.

        Returns
        -------
        dict
            Combined stage report containing metrics from both reports.
        """
        combined_stage_report: dict[str, StageReport] = {}
        for stage_name, stage_report in previous_stage_report.items():
            new_stage_report = StageReport(name=stage_name, metrics={})
            if stage_name not in current_stage_report:
                current_stage_report[stage_name] = StageReport(
                    name=stage_name, metrics={}
                )
            for metric_name, instance_report_dict in stage_report.metrics.items():
                if metric_name not in current_stage_report[stage_name].metrics:
                    current_stage_report[stage_name].metrics[metric_name] = {}
                new_stage_report.metrics[metric_name] = {
                    **current_stage_report[stage_name].metrics[metric_name],
                    **instance_report_dict,
                }
            combined_stage_report[stage_name] = new_stage_report
        return combined_stage_report

    def process_previous_report(self, previous_report_dict: dict) -> None:
        """
        Process and incorporate a previous report.

        Parameters
        ----------
        previous_report_dict : dict
            Dictionary representation of a previous report.

        Notes
        -----
        - Service quality, data quality, and ML inference reports are combined with the current report.
        """
        previous_report = GeneralMlInferenceReport(**previous_report_dict)
        self.previous_report.append(previous_report)

        self.report.service = self.combine_stage_report(
            self.report.service, previous_report.service
        )
        self.report.data = self.combine_stage_report(
            self.report.data, previous_report.data
        )
        self.report.ml_inference |= previous_report.ml_inference

    def observe_metric(
        self, report_type: ReportTypeEnum, stage: str, metric: Metric
    ) -> None:
        """
        Observe and record a metric.

        Parameters
        ----------
        report_type : ReportTypeEnum
            The type of report being generated.
        stage : str
            The stage of the process in which the metric is recorded.
        metric : Metric
            The metric to be recorded.

        Raises
        ------
        ValueError
            If the stage name is empty or the report type is not handled.
        """
        if stage == "":
            raise ValueError("Stage name can't be empty")

        report_dict = None
        if report_type == ReportTypeEnum.service:
            if stage not in self.report.service:
                self.report.service[stage] = StageReport(name=stage, metrics={})
            report_dict = self.report.service[stage].metrics
        elif report_type == ReportTypeEnum.data:
            if stage not in self.report.data:
                self.report.data[stage] = StageReport(name=stage, metrics={})
            report_dict = self.report.data[stage].metrics
        else:
            raise ValueError(f"Can't handle report type {report_type}")

        if metric.metric_name not in report_dict:
            report_dict[metric.metric_name] = {}

        report_dict[metric.metric_name][UUID(self.client_config.instance_id)] = metric

    def observe_inference(self, inference_value: Any) -> None:
        """
        Observe and record inference data.

        Parameters
        ----------
        inference_value : Any
            The value of the inference to be recorded.

        Notes
        -----
        - Raises a warning if inference data already exists for the current instance.
        """
        instance_id = UUID(self.client_config.instance_id)

        if instance_id in self.report.ml_inference:
            raise RuntimeWarning(
                "Inference existed, will override the existing inference"
            )

        self.report.ml_inference[instance_id] = InferenceInstance(
            inference_id=uuid4(),
            instance_id=instance_id,
            functionality=self.client_config.functionality,
            prediction=inference_value,
        )

    def observe_inference_metric(self, metric: Metric) -> None:
        """
        Observe and record an inference-specific metric.

        Parameters
        ----------
        metric : Metric
            The inference-specific metric to be recorded.
        """
        instance_id = UUID(self.client_config.instance_id)

        if instance_id in self.report.ml_inference:
            self.report.ml_inference[instance_id].metrics.append(metric)
        else:
            self.report.ml_inference[instance_id] = InferenceInstance(
                inference_id=uuid4(),
                instance_id=instance_id,
                functionality=self.client_config.functionality,
                metrics=[metric],
            )

    def generate_report(
        self, reset: bool = True, corr_id: Optional[str] = None
    ) -> BaseReport:
        """
        Generate the report and optionally reset the current report state.

        Parameters
        ----------
        reset : bool, optional
            Whether to reset the report state after generating the report, default is True.
        corr_id : Optional[str], optional
            Correlation ID for the report, default is None.

        Returns
        -------
        BaseReport
            The generated report.

        Notes
        -----
        - Adds metadata such as client configuration, timestamp, and runtime to the report.
        - Deep copies the current state of the report before optionally resetting it.
        """
        self.report.metadata["client_config"] = copy.deepcopy(self.client_config)
        self.report.metadata["timestamp"] = time.time()
        if corr_id is not None:
            self.report.metadata["corr_id"] = corr_id
        self.report.metadata["runtime"] = (
            self.report.metadata["timestamp"] - self.init_time
        )

        report = copy.deepcopy(self.report)
        if reset:
            self.reset()
        return report



import inspect
from enum import Enum, EnumMeta

QOA_ATTRIBUTES_VERSION = "v0.3"
QOA_ATTRIBUTES_NAME = "qoa4ml-attributes"


class MetaEnum(EnumMeta):
    def __new__(cls, clsname, bases, classdict):
        cls = super().__new__(cls, clsname, bases, classdict)

        source = inspect.getsource(cls)
        docstrings = source.split('"""')[-2::-2]

        for member_name, doc_str in zip(reversed(cls._member_names_), docstrings):
            enum_member = getattr(cls, member_name)
            enum_member.__doc__ = doc_str.strip()

        return cls


class QoAttribute(str, Enum, metaclass=MetaEnum):
    pass


class DataQualityEnum(QoAttribute):
    ACCURACY = "accuracy"
    """The ratio between correct and total data the service received (%)"""

    COMPLETENESS = "completeness"
    """The ratio between received and expected number of data attributes sent to the service"""

    TOTAL_ERRORS = "total_errors"
    """Total number of errors as given by the user in the data"""

    ERROR_RATIOS = "error_ratios"
    """Ratio of errors in the data"""

    DUPLICATE_RATIO = "duplicate_ratio"
    """Ratio of duplicate entries in the data"""

    TOTAL_DUPLICATE = "total_duplicate"
    """Total number of duplicate entries in the data"""

    NULL_COUNT = "null_count"
    """Count of null or undefined entries in the data"""

    NULL_CORRELATIONS = "null_correlations"
    """Correlation between null or undefined entries in the data"""

    TOTAL_VALID = "total_valid"
    """Total number of valid entries in the data"""

    TOTAL_NONE = "total_none"
    """Total number of none or empty entries in the data"""

    NONE_RATIO = "none_ratio"
    """Ratio of none or empty entries in the data"""


class MLModelQualityEnum(QoAttribute):
    AUC = "auc"
    """The measure of the ability of a classifier to distinguish between classes and is used as a summary of the ROC curve"""

    ACCURACY = "accuracy"
    """Can be measured in different ways such as confidence score in classification models"""

    MSE = "mse"
    """Mean square error used for regression models"""

    PRECISION = "precision"
    """The fraction of true positive responses over total number of positive responses"""

    RECALL = "recall"
    """The fraction of true positive responses over total number of correct responses"""


class ServiceQualityEnum(QoAttribute):
    AVAILABILITY = "availability"
    """The ratio between up time and down time of the service (%)"""

    RELIABILITY = "reliability"
    """The ratio between correct and total service responses (%)"""

    RESPONSE_TIME = "response_time"
    """The response time of each microservice, measured by the time period between receiving and replying the request (s)"""



import glob
import json
import logging
import os
import pathlib
import re
import shlex
import subprocess
import sys
import time
import traceback
from threading import Thread
from typing import Any, Optional

import numpy as np
import psutil
import yaml

from .logger import qoa_logger


def make_folder(temp_path: str) -> bool:
    """
    Create a folder if it doesn't already exist.

    Parameters
    ----------
    temp_path : str
        The path of the folder to be created.

    Returns
    -------
    bool
        True if the folder exists or is created successfully, False otherwise.

    Notes
    -----
    If the folder already exists, nothing is done.
    """
    try:
        if not os.path.exists(temp_path):
            os.makedirs(temp_path)
        return True
    except Exception:
        return False


def get_cgroup_version() -> str:
    """
    Retrieve the current cgroup version.

    Returns
    -------
    str
        The cgroup version ("v1" or "v2").

    Notes
    -----
    Uses subprocess to execute the `mount` command and grep for cgroup version.
    """
    proc1 = subprocess.Popen("mount", stdout=subprocess.PIPE)
    proc2 = subprocess.Popen(
        shlex.split("grep cgroup"),
        stdin=proc1.stdout,
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
    )
    if proc1.stdout:
        proc1.stdout.close()
    out, _ = proc2.communicate()
    return "v2" if "cgroup2" in out.decode() else "v1"


if get_cgroup_version() == "v2":
    CGROUP_VERSION = "v2"
else:
    CGROUP_VERSION = "v1"


def set_logger_level(logging_level: int) -> None:
    """
    Set the logging level for the application logger.

    Parameters
    ----------
    logging_level : int
        The desired logging level:
        0 - NOTSET
        1 - DEBUG
        2 - INFO
        3 - WARNING
        4 - ERROR
        5 - CRITICAL

    Raises
    ------
    ValueError
        If the logging level is not between 0 and 5.
    """
    log_levels = [
        logging.NOTSET,
        logging.DEBUG,
        logging.INFO,
        logging.WARNING,
        logging.ERROR,
        logging.CRITICAL,
    ]

    if not 0 <= logging_level < len(log_levels):
        raise ValueError(f"Error logging level {logging_level}")

    qoa_logger.setLevel(log_levels[logging_level])


def load_config(file_path: str) -> Optional[dict]:
    """
    Load a configuration file.

    Parameters
    ----------
    file_path : str
        The path to the configuration file.

    Returns
    -------
    dict
        The loaded configuration dictionary.

    Notes
    -----
    Supports JSON and YAML file formats. Logs a warning if the format is unsupported.
    """
    try:
        with open(file_path) as f:
            if "json" in file_path:
                return json.load(f)
            elif "yaml" in file_path or "yml" in file_path:
                return yaml.safe_load(f)
            else:
                qoa_logger.warning("Unsupported format")
    except Exception:
        qoa_logger.error("Unable to load configuration")
        return None


def to_json(file_path: str, conf: dict) -> None:
    """
    Save a configuration to a JSON file.

    Parameters
    ----------
    file_path : str
        The path to the file where the configuration should be saved.
    conf : dict
        The configuration dictionary to save.
    """
    with open(file_path, "w") as f:
        json.dump(conf, f)


def to_yaml(file_path: str, conf: dict) -> None:
    """
    Save a configuration to a YAML file.

    Parameters
    ----------
    file_path : str
        The path to the file where the configuration should be saved.
    conf : dict
        The configuration dictionary to save.
    """
    with open(file_path, "w") as f:
        yaml.dump(conf, f)


def get_sys_cpu() -> dict:
    """
    Retrieve system CPU statistics and times.

    Returns
    -------
    dict
        Dictionary containing CPU stats and times.

    Notes
    -----
    Uses psutil to retrieve both CPU stats and times.
    """
    stats = psutil.cpu_stats()
    cpu_time = psutil.cpu_times()
    return {key: getattr(stats, key) for key in stats._fields} | {
        key: getattr(cpu_time, key) for key in cpu_time._fields
    }


def get_sys_cpu_util() -> dict:
    """
    Retrieve system CPU utilization for each core.

    Returns
    -------
    dict
        Dictionary containing CPU utilization for each core.
    """
    core_utils = psutil.cpu_percent(percpu=True)
    return {
        f"core_{core_num}": core_util for core_num, core_util in enumerate(core_utils)
    }


def get_sys_cpu_metadata() -> dict:
    """
    Retrieve metadata information about the system CPU.

    Returns
    -------
    dict
        Dictionary containing CPU frequency and thread count.
    """
    cpu_freq = psutil.cpu_freq()
    return {
        "frequency": {"value": cpu_freq.max / 1000, "unit": "GHz"},
        "thread": psutil.cpu_count(logical=True),
    }


def get_sys_mem() -> dict:
    """
    Retrieve system memory statistics.

    Returns
    -------
    dict
        Dictionary containing memory stats.
    """
    stats = psutil.virtual_memory()
    return {key: getattr(stats, key) for key in stats._fields}


def get_sys_net() -> dict:
    """
    Retrieve system network I/O statistics.

    Returns
    -------
    dict
        Dictionary containing network I/O stats.
    """
    net = psutil.net_io_counters()
    return {key: getattr(net, key) for key in net._fields}


def report_proc_cpu(process: psutil.Process) -> dict:
    """
    Retrieve CPU usage statistics for a given process.

    Parameters
    ----------
    process : psutil.Process
        The process to retrieve CPU stats for.

    Returns
    -------
    dict
        Dictionary containing CPU stats for the process.
    """
    cpu_time = process.cpu_times()
    context = process.num_ctx_switches()
    return (
        {key: getattr(cpu_time, key) for key in cpu_time._fields}
        | {key: getattr(context, key) for key in context._fields}
        | {"num_thread": process.num_threads()}
    )


def report_proc_child_cpu(process: psutil.Process) -> dict:
    """
    Retrieve CPU usage statistics for a given process and its children.

    Parameters
    ----------
    process : psutil.Process
        The process to retrieve CPU stats for.

    Returns
    -------
    dict
        Dictionary containing CPU stats for the process and its children.

    Notes
    -----
    This function can be time-consuming as it recursively evaluates all child processes.
    """
    child_processes = process.children(recursive=True)
    child_processes_cpu = {
        f"child_{id}": float(
            child_proc.cpu_times().user + child_proc.cpu_times().system
        )
        for id, child_proc in enumerate(child_processes)
    }

    process_cpu_time = process.cpu_times()
    main_process = float(
        process_cpu_time.user
        + process_cpu_time.system
        + process_cpu_time.children_user
        + process_cpu_time.children_system
    )
    total_cpu_usage = sum(child_processes_cpu.values()) + main_process

    return {
        "child_process": len(child_processes),
        "value": child_processes_cpu,
        "main_process": main_process,
        "total": total_cpu_usage,
        "unit": "cputime",
    }


def get_proc_cpu(pid: Optional[int] = None) -> dict:
    """
    Retrieve CPU usage statistics for a given process and its children.

    Parameters
    ----------
    pid : int, optional
        The process ID to retrieve CPU stats for. If None, uses the current process ID.

    Returns
    -------
    dict
        Dictionary containing CPU stats for the process and its children.

    Notes
    -----
    - The main process's stats are keyed by its PID.
    - Each child process's stats are keyed by its PID with a "c" suffix.
    """
    if pid is None:
        pid = os.getpid()
    process = psutil.Process(pid)
    child_list = process.children()
    info = {}
    info[pid] = report_proc_cpu(process)

    for child in child_list:
        info[f"{child.pid}c"] = report_proc_cpu(child)
    return info


def report_proc_mem(process: psutil.Process) -> dict:
    """
    Retrieve memory usage statistics for a given process.

    Parameters
    ----------
    process : psutil.Process
        The process to retrieve memory stats for.

    Returns
    -------
    dict
        Dictionary containing memory stats for the process.
    """
    mem_info = process.memory_info()
    return {key: getattr(mem_info, key) for key in mem_info._fields}


def get_proc_mem(pid: Optional[int] = None) -> dict:
    """
    Retrieve memory usage statistics for a given process and its children.

    Parameters
    ----------
    pid : int, optional
        The process ID to retrieve memory stats for. If None, uses the current process ID.

    Returns
    -------
    dict
        Dictionary containing memory stats for the process and its children.

    Notes
    -----
    - The main process's stats are keyed by its PID.
    - Each child process's stats are keyed by its PID with a "c" suffix.
    """
    if pid is None:
        pid = os.getpid()
    process = psutil.Process(pid)
    child_list = process.children()
    info = {}
    info[pid] = report_proc_mem(process)

    for child in child_list:
        info[f"{child.pid}c"] = report_proc_mem(child)
    return info


def convert_to_gbyte(value: float) -> float:
    """
    Convert a value from bytes to gigabytes.

    Parameters
    ----------
    value : float
        The value in bytes to be converted.

    Returns
    -------
    float
        The converted value in gigabytes.
    """
    return value / 1024.0 / 1024.0 / 1024.0


def convert_to_mbyte(value: float) -> float:
    """
    Convert a value from bytes to megabytes.

    Parameters
    ----------
    value : float
        The value in bytes to be converted.

    Returns
    -------
    float
        The converted value in megabytes.
    """
    return value / 1024.0 / 1024.0


def convert_to_kbyte(value: float) -> float:
    """
    Convert a value from bytes to kilobytes.

    Parameters
    ----------
    value : float
        The value in bytes to be converted.

    Returns
    -------
    float
        The converted value in kilobytes.
    """
    return value / 1024.0


###################### SYSTEM REPORT ######################

sys_monitor_flag = False
process_monitor_flag = False
doc_monitor_flag = False


def system_report(
    client, interval: int, to_mb: bool = True, to_gb: bool = False, to_kb: bool = False
) -> None:
    """
    Generate a system report and send it to the client at regular intervals.

    Parameters
    ----------
    client : Any
        The client to send the report to.
    interval : int
        The time interval in seconds between reports.
    to_mb : bool, optional
        Convert network I/O values to megabytes, default is True.
    to_gb : bool, optional
        Convert network I/O values to gigabytes, default is False.
    to_kb : bool, optional
        Convert network I/O values to kilobytes, default is False.

    Notes
    -----
    - Collects CPU, memory, and network I/O statistics.
    - Logs errors if any occur during data collection or report sending.
    - Sleeps for the specified interval between reports.
    """
    report = {}
    last_net_value = {"sent": 0, "receive": 0}
    while sys_monitor_flag:
        try:
            report["sys_cpu_stats"] = get_sys_cpu()
        except Exception as e:
            qoa_logger.error(f"Error {type(e)} in report CPU stat: {e}")
            traceback.print_exception(*sys.exc_info())
        try:
            report["sys_mem_stats"] = get_sys_mem()
        except Exception as e:
            qoa_logger.error(f"Error {type(e)} in report memory stat: {e}")
            traceback.print_exception(*sys.exc_info())
        try:
            report["sys_net_stats"] = get_sys_net()
            sent = 0
            receive = 0
            if to_mb:
                sent = convert_to_mbyte(psutil.net_io_counters().bytes_sent)
                receive = convert_to_mbyte(psutil.net_io_counters().bytes_recv)
            elif to_gb:
                sent = convert_to_gbyte(psutil.net_io_counters().bytes_sent)
                receive = convert_to_gbyte(psutil.net_io_counters().bytes_recv)
            elif to_kb:
                sent = convert_to_kbyte(psutil.net_io_counters().bytes_sent)
                receive = convert_to_kbyte(psutil.net_io_counters().bytes_recv)
            else:
                sent = psutil.net_io_counters().bytes_sent
                receive = psutil.net_io_counters().bytes_recv

            curr_net_value = {"sent": sent, "receive": receive}
            report["sys_net_send"] = curr_net_value["sent"] - last_net_value["sent"]
            report["sys_net_receive"] = (
                curr_net_value["receive"] - last_net_value["receive"]
            )
            last_net_value = curr_net_value.copy()
        except Exception as e:
            qoa_logger.error(f"Error {type(e)} in report network stat: {e}")
            traceback.print_exception(*sys.exc_info())
        try:
            client.report(report=report)
        except Exception as e:
            qoa_logger.error(f"Error {type(e)} in sent system report: {e}")
            traceback.print_exception(*sys.exc_info())
        time.sleep(interval)


def sys_monitor(client, interval: int) -> None:
    """
    Start monitoring system reports.

    Parameters
    ----------
    client : Any
        The client to send the report to.
    interval : int
        The time interval in seconds between reports.

    Notes
    -----
    - Starts a new thread to generate system reports at regular intervals.
    """
    sub_thread = Thread(target=system_report, args=(client, interval))
    sub_thread.start()


###################### PROCESS REPORT ######################

# def process_report(client, interval:int, pid:int = None):
#     report = {}
#     while procMonitorFlag:
#         try:
#             report["proc_cpu_stats"] = get_proc_cpu()
#         except Exception as e:
#             qoaLogger.error("Error {} in report process cpu stat: {}".format(type(e),e.__traceback__))
#             traceback.print_exception(*sys.exc_info())
#         try:
#             report["proc_mem_stats"] = get_proc_mem()
#         except Exception as e:
#             qoaLogger.error("Error {} in report process memory stat: {}".format(type(e),e.__traceback__))
#             traceback.print_exception(*sys.exc_info())
#         try:
#             client.report(report=report)
#         except Exception as e:
#             qoaLogger.error("Error {} in sent process report: {}".format(type(e),e.__traceback__))
#             traceback.print_exception(*sys.exc_info())
#         time.sleep(interval)


# def process_monitor(client, interval:int, pid:int = None):
#     if (pid == None):
#         pid = os.getpid()
#     sub_thread = Thread(target=process_report, args=(client, interval, pid))
#     sub_thread.start()


###################### DOCKER REPORT ######################


def get_cpu_stat(stats: dict, key: str) -> float:
    """
    Retrieve CPU usage statistics from Docker stats.

    Parameters
    ----------
    stats : dict
        The Docker stats dictionary.
    key : str
        The key indicating the type of CPU statistic (e.g., "percentage").

    Returns
    -------
    float
        The CPU usage percentage, or -1 if the key is not recognized.

    Notes
    -----
    - Calculates the CPU usage percentage based on the difference between the current and previous CPU usage.
    """
    if key == "percentage":
        usage_delta = (
            stats["cpu_stats"]["cpu_usage"]["total_usage"]
            - stats["precpu_stats"]["cpu_usage"]["total_usage"]
        )
        system_delta = (
            stats["cpu_stats"]["system_cpu_usage"]
            - stats["precpu_stats"]["system_cpu_usage"]
        )
        len_cpu = stats["cpu_stats"]["online_cpus"]
        percentage = (usage_delta / system_delta) * len_cpu * 100
        return round(percentage, 2)
    return -1


def get_mem_stat(stats: dict, key: str) -> int:
    """
    Retrieve memory usage statistics from Docker stats.

    Parameters
    ----------
    stats : dict
        The Docker stats dictionary.
    key : str
        The key indicating the type of memory statistic (e.g., "used").

    Returns
    -------
    int
        The memory usage in bytes, or -1 if the key is not recognized.
    """
    if key == "used":
        return stats["memory_stats"]["usage"]
    return -1


def merge_report(f_report: dict, i_report: dict, prio: bool = True) -> dict:
    """
    Merge two report dictionaries.

    Parameters
    ----------
    f_report : dict
        The first report dictionary.
    i_report : dict
        The second report dictionary.
    prio : bool, optional
        Flag to determine which report takes priority in case of conflict, default is True.

    Returns
    -------
    dict
        The merged report dictionary.

    Notes
    -----
    - If both reports are dictionaries, merges them recursively.
    - If there is a conflict and prio is True, the value from f_report is used; otherwise, the value from i_report is used.
    """
    try:
        if isinstance(f_report, dict) and isinstance(i_report, dict):
            key_list = tuple(f_report.keys())
            for key in key_list:
                if key in i_report:
                    f_report[key] = merge_report(f_report[key], i_report[key], prio)
                    i_report.pop(key)
            f_report.update(i_report)
        elif f_report != i_report:
            return f_report if prio else i_report
    except Exception as e:
        qoa_logger.error(f"Error {type(e)} in merge_report: {e.__traceback__}")
        traceback.print_exception(*sys.exc_info())
    return f_report


def get_dict_at(dictionary: dict, i: int = 0):
    """
    Retrieve the key-value pair at a specific index in a dictionary.

    Parameters
    ----------
    dictionary : dict
        The dictionary from which to retrieve the key-value pair.
    i : int, optional
        The index of the key-value pair to retrieve, default is 0.

    Returns
    -------
    tuple
        A tuple containing the key and the value at the specified index.

    Raises
    ------
    IndexError
        If the index is out of range.

    Notes
    -----
    - Logs an error and prints the exception traceback if an error occurs.
    """
    try:
        keys = list(dictionary.keys())
        return keys[i], dictionary[keys[i]]
    except Exception as e:
        qoa_logger.error(f"Error {type(e)} in get_dict_at: {e.__traceback__}")
        traceback.print_exception(*sys.exc_info())


def get_file_dir(file: str, to_string: bool = True):
    """
    Get the directory of a file.

    Parameters
    ----------
    file : str
        The file path.
    to_string : bool, optional
        Flag to return the directory as a string, default is True.

    Returns
    -------
    str or pathlib.Path
        The directory of the file as a string or Path object.
    """
    current_dir = pathlib.Path(file).parent.absolute()
    return str(current_dir) if to_string else current_dir


def get_parent_dir(file: str, parent_level: int = 1, to_string: bool = True):
    """
    Get the parent directory of a file by a specified number of levels.

    Parameters
    ----------
    file : str
        The file path.
    parent_level : int, optional
        The number of levels up to retrieve the parent directory, default is 1.
    to_string : bool, optional
        Flag to return the directory as a string, default is True.

    Returns
    -------
    str or pathlib.Path
        The parent directory of the file as a string or Path object.
    """
    current_dir = get_file_dir(file=file, to_string=False)
    for _ in range(parent_level):
        current_dir = current_dir.parent.absolute()
    return str(current_dir) if to_string else current_dir


def is_numpyarray(obj: Any) -> bool:
    """
    Check if an object is a NumPy array.

    Parameters
    ----------
    obj : Any
        The object to check.

    Returns
    -------
    bool
        True if the object is a NumPy array, False otherwise.
    """
    return isinstance(obj, np.ndarray)


def get_process_allowed_cpus() -> list[int]:
    """
    Retrieve the list of CPU cores available to the process.

    Returns
    -------
    list[int]
        A list of CPU core indices.

    Notes
    -----
    - Uses the call process's PID (0) to get the CPU affinity.
    """
    pid = 0
    affinity = os.sched_getaffinity(pid)
    return list(affinity)


def get_process_allowed_memory() -> Optional[float]:
    """
    Retrieve the memory limit allowed to the process.

    Returns
    -------
    Optional[float]
        The memory limit in bytes, or None if unable to retrieve.

    Notes
    -----
    - Supports both cgroup v1 and v2 formats to get the memory limit.
    """
    if CGROUP_VERSION == "v1":
        with open("/proc/self/cgroup") as file:
            for line in file:
                parts = line.strip().split(":")
                if len(parts) == 3 and parts[1] == "memory":
                    cgroup_path = parts[2]
                    memory_limit_file = re.sub(r"/task_\d+", "", cgroup_path)

                    number_of_tasks = len(
                        glob.glob(f"/sys/fs/cgroup/memory{memory_limit_file}/task_*")
                    )

                    with open(
                        f"/sys/fs/cgroup/memory{memory_limit_file}/memory.limit_in_bytes"
                    ) as limit_file:
                        memory_limit_str = limit_file.read().strip()
                        try:
                            memory_limit_int = int(memory_limit_str)
                            return memory_limit_int / number_of_tasks
                        except ValueError:
                            return memory_limit_str
            return None
    else:
        with open("/proc/self/cgroup") as file:
            for line in file:
                parts = line.strip().split(":")
                cgroup_path = parts[2]
                pattern = r"/task_\d+"
                cgroup_path = re.sub(pattern, "", cgroup_path)
                with open(f"/sys/fs/cgroup{cgroup_path}/memory.max") as limit_file:
                    number_of_tasks = len(
                        glob.glob(f"/sys/fs/cgroup{cgroup_path}/task_*")
                    )
                    memory_limit_str = limit_file.read().strip()
                    try:
                        memory_limit_int = int(memory_limit_str)
                        return memory_limit_int / number_of_tasks
                    except ValueError:
                        return memory_limit_str
            return None



import json
import time

import docker

from qoa4ml.reports.resources_report_model import DockerReport

from ..config.configs import ClientInfo, DockerProbeConfig
from ..connector.base_connector import BaseConnector
from ..utils.docker_util import get_docker_stats
from ..utils.logger import qoa_logger
from .probe import Probe


class DockerMonitoringProbe(Probe):
    """
    DockerMonitoringProbe is responsible for monitoring Docker containers and creating reports.

    Parameters
    ----------
    config : DockerProbeConfig
        Configuration settings for the Docker monitoring probe.
    connector : BaseConnector
        Connector to send the report data.
    client_info : ClientInfo
        Information about the client.

    Attributes
    ----------
    config : DockerProbeConfig
        The Docker monitoring probe configuration.
    obs_service_url : str
        The URL of the observation service, if registration is required.
    docker_client : docker.DockerClient
        The Docker client for communicating with Docker API.

    Methods
    -------
    create_report() -> str
        Create a report based on Docker container statistics.
    """

    def __init__(
        self,
        config: DockerProbeConfig,
        connector: BaseConnector,
        client_info: ClientInfo,
    ) -> None:
        """
        Initialize an instance of DockerMonitoringProbe.

        Parameters
        ----------
        config : DockerProbeConfig
            Configuration settings for the Docker monitoring probe.
        connector : BaseConnector
            Connector to send the report data.
        client_info : ClientInfo
            Information about the client.
        """
        super().__init__(config, connector, client_info)
        self.config = config
        if self.config.require_register:
            self.obs_service_url = self.config.obs_service_url
        self.docker_client = docker.from_env()

    def create_report(self) -> str:
        """
        Create a report based on Docker container statistics.

        Returns
        -------
        str
            JSON-encoded report containing Docker container statistics.

        Notes
        -----
        - This method collects statistics for the specified Docker containers.
        - If the report dictionary is empty, it adds a 2-second delay to prevent fast looping.
        - In case of a RuntimeError, an error message is returned in a JSON format.
        """
        try:
            reports = get_docker_stats(self.docker_client, self.config.container_list)
            docker_report = DockerReport(
                metadata=self.client_info,
                timestamp=time.time(),
                container_reports=reports,
            )
            reports_dict = docker_report.model_dump()
            if not reports_dict:
                time.sleep(2)
            return json.dumps(reports_dict)
        except RuntimeError:
            qoa_logger.exception(
                "RuntimeError occurred, possibly due to running in the background!"
            )
        return json.dumps({"error": "RuntimeError"})



from abc import ABC, abstractmethod


class BaseConnector(ABC):
    @abstractmethod
    def send_report(self, body_message: str):
        pass

    def check_connection(self) -> bool:
        return True
